<?php

namespace Pinpoint\BlankProduct\Model;

use Magento\Catalog\Model\ProductFactory;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\CatalogInventory\Api\StockConfigurationInterface;
use Magento\CatalogInventory\Model\ResourceModel\QtyCounterInterface;
use Magento\CatalogInventory\Model\ResourceModel\Stock as ResourceStock;
use Magento\CatalogInventory\Model\Spi\StockRegistryProviderInterface;
use Magento\CatalogInventory\Model\StockState;
use \Magento\Framework\Registry;

/**
 * Class StockManagement
 * @package Pinpoint\BlankProduct\Model
 */
class StockManagement extends \Magento\CatalogInventory\Model\StockManagement
{
    /**
     * @var QtyCounterInterface
     */
    protected $_qtyCounter;

    /**
     * @var ProductFactory
     */
    protected $_productFactory;

    /**
     * @var Registry
     */
    protected $_coreRegistry;

    /**
     * @var \Magento\Checkout\Model\Session
     */
    protected $checkoutSession;

    /**
     * @var \Magento\Backend\Model\Session\Quote
     */
    protected $backendQuoteSession;

    /**
     * StockManagement constructor.
     * @param ResourceStock $stockResource
     * @param StockRegistryProviderInterface $stockRegistryProvider
     * @param StockState $stockState
     * @param StockConfigurationInterface $stockConfiguration
     * @param ProductRepositoryInterface $productRepository
     * @param QtyCounterInterface $qtyCounter
     * @param ProductFactory $productFactory
     * @param Registry $coreRegistry
     * @param \Magento\Checkout\Model\Session $checkoutSession
     * @param \Magento\Backend\Model\Session\Quote $backendQuoteSession
     */
    public function __construct(
        ResourceStock $stockResource,
        StockRegistryProviderInterface $stockRegistryProvider,
        StockState $stockState,
        StockConfigurationInterface $stockConfiguration,
        ProductRepositoryInterface $productRepository,
        QtyCounterInterface $qtyCounter,
        ProductFactory $productFactory,
        Registry $coreRegistry,
        \Magento\Checkout\Model\Session $checkoutSession,
        \Magento\Backend\Model\Session\Quote $backendQuoteSession
    ) {
        $this->_qtyCounter = $qtyCounter;
        $this->_productFactory = $productFactory;
        $this->_coreRegistry = $coreRegistry;
        $this->checkoutSession = $checkoutSession;
        $this->backendQuoteSession = $backendQuoteSession;

        parent::__construct(
            $stockResource,
            $stockRegistryProvider,
            $stockState,
            $stockConfiguration,
            $productRepository,
            $qtyCounter
        );
    }

    /**
     * @param string[] $items
     * @param null $websiteId
     * @return array
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function registerProductsSale($items, $websiteId = null)
    {
        $websiteId = $this->stockConfiguration->getDefaultScopeId();

        $this->getResource()->beginTransaction();
        $lockedItems = $this->getResource()->lockProductsStock(array_keys($items), $websiteId);
        $fullSaveItems = $registeredItems = [];
        $blankProducts = [];

        $usedBlankProducts = [];

        foreach ($lockedItems as $lockedItemRecord) {
            $productId = $lockedItemRecord['product_id'];
            $orderedQty = $items[$productId];

            /** @var \Magento\CatalogInventory\Api\Data\StockItemInterface $stockItem */
            $stockItem = $this->stockRegistryProvider->getStockItem($productId, $websiteId);
            $canSubtractQty = $stockItem->getItemId() && $this->canSubtractQty($stockItem);
            if (!$canSubtractQty || !$this->stockConfiguration->isQty($lockedItemRecord['type_id'])) {
                continue;
            }

            $neededBlankProductsQty = $orderedQty;

            for ($i = 1; $i <= $orderedQty; $i++) {
                $result = $this->stockState->checkQty($productId, $i, $stockItem->getWebsiteId());
                if ($result) {
                    $neededBlankProductsQty -= 1;
                }
            }

            $originalQty = $orderedQty - $neededBlankProductsQty;

            if ($neededBlankProductsQty) {
                $_product =$this->_productFactory->create()->load($productId);
                if ($_product->getCustomAttribute('blank_products') &&
                    trim($_product->getCustomAttribute('blank_products')->getValue()) != '') {
                    $products = explode(",", $_product->getCustomAttribute('blank_products')->getValue());
                    foreach ($products as $blankProduct) {
                        $productData = explode("|", $blankProduct);
                        $_blankProduct = $this->_productFactory->create()->loadByAttribute('sku', $productData[0]);
                        $blankProducts[] = [
                            'product' => $_blankProduct,
                            'stockItem' => $this->stockRegistryProvider->getStockItem($_blankProduct->getId(), $websiteId),
                            'qty_required' => $productData[1]
                        ];
                    }
                }
            }


            if ($quote = $this->checkoutSession->getQuote()) {
                if (is_null($quote->getItems())) {
                    $quote = $this->backendQuoteSession->getQuote();
                }

                foreach ($quote->getItems() as $item) {
                    if ($item->getProductId() == $productId) {
                        preg_match_all('/(?<=\#)(.*?)(?=\#)/', $item->getSku(), $skus);
                        $skus = array_filter($skus[0], function ($value) { return $value !== '-'; });

                        if (is_array($skus) && count($skus)) {
                            foreach ($skus as $blankSku) {
                                $_blankProduct = $this->_productFactory->create()->loadByAttribute('sku', $blankSku);
                                $blankProducts[] = [
                                    'product' => $_blankProduct,
                                    'stockItem' => $this->stockRegistryProvider->getStockItem($_blankProduct->getId(), $websiteId),
                                    'qty_required' => $neededBlankProductsQty
                                ];
                            }
                        }
                    }
                }
            }

            if (!$stockItem->hasAdminArea()
                && !$this->stockState->checkQty($productId, $originalQty, $stockItem->getWebsiteId())
            ) {
                $this->getResource()->commit();
                throw new \Magento\Framework\Exception\LocalizedException(
                    __('Not all of your products are available in the requested quantity.')
                );
            }

            if ($this->canSubtractQty($stockItem)) {
                $stockItem->setQty($stockItem->getQty() - $originalQty);
            }

            if ($neededBlankProductsQty) {
                foreach ($blankProducts as $blankProduct) {
                    $requiredBlankQty = $neededBlankProductsQty * $blankProduct['qty_required'];
                    if (!$blankProduct['stockItem']->hasAdminArea()
                        && !$this->stockState->checkQty($blankProduct['product']->getId(), $requiredBlankQty, $blankProduct['stockItem']->getWebsiteId())
                    ) {
                        $this->getResource()->commit();
                        throw new \Magento\Framework\Exception\LocalizedException(
                            __('Not all of your products are available in the requested quantity.')
                        );
                    }

                    if ($this->canSubtractQty($blankProduct['stockItem'])) {
                        $blankProduct['stockItem']->setQty($blankProduct['stockItem']->getQty() - $requiredBlankQty);
                        $usedBlankProducts[$productId][] = [$blankProduct['product']->getId() => $requiredBlankQty];
                    }
                }
            }


            $registeredItems[$productId] = $originalQty;
            if (!$this->stockState->verifyStock($productId, $stockItem->getWebsiteId())
                || $this->stockState->verifyNotification(
                    $productId,
                    $stockItem->getWebsiteId()
                )
            ) {
                $fullSaveItems[] = $stockItem;
            }

            if ($neededBlankProductsQty && count($blankProducts)) {
                foreach ($blankProducts as $blankProduct) {
                    $registeredItems[$blankProduct['product']->getId()] = $neededBlankProductsQty * $blankProduct['qty_required'];
                    if (!$this->stockState->verifyStock($productId, $blankProduct['stockItem']->getWebsiteId())
                        || $this->stockState->verifyNotification(
                            $blankProduct['product']->getId(),
                            $blankProduct['stockItem']->getWebsiteId()
                        )
                    ) {
                        $fullSaveItems[] = $blankProduct['stockItem'];
                    }
                }
            }

            if ($quote = $this->checkoutSession->getQuote()) {
                if (is_null($quote->getItems())) {
                    $quote = $this->backendQuoteSession->getQuote();
                }
                if (count($quote->getItems())) {
                    foreach ($quote->getItems() as $item) {
                        if (isset($usedBlankProducts[$item->getProductId()])) {
                            $item->setData('blank_products', \json_encode($usedBlankProducts[$productId]))->save();
                        }
                    }
                }
            }
        }

        $this->_qtyCounter->correctItemsQty($registeredItems, $websiteId, '-');
        $this->getResource()->commit();

        return $fullSaveItems;
    }

}